/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2002 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2001 - 2004 John Weidman (jweidman@slip.net)

  (c) Copyright 2002 - 2004 Brad Jorsch (anomie@users.sourceforge.net),
                            funkyass (funkyass@spam.shaw.ca),
                            Joel Yliluoma (http://iki.fi/bisqwit/)
                            Kris Bleakley (codeviolation@hotmail.com),
                            Matthew Kendora,
                            Nach (n-a-c-h@users.sourceforge.net),
                            Peter Bortas (peter@bortas.org) and
                            zones (kasumitokoduck@yahoo.com)

  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and Nach

  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2004 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman, neviksti (neviksti@hotmail.com),
                            Kris Bleakley, Andreas Naive

  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2004 zsKnight, pagefault (pagefault@zsnes.com) and
                            Kris Bleakley
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-DD1 C emulator code
  (c) Copyright 2003 Brad Jorsch with research by
                     Andreas Naive and John Weidman
 
  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  ST010 C++ emulator code
  (c) Copyright 2003 Feather, Kris Bleakley, John Weidman and Matthew Kendora

  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar, Gary Henderson and John Weidman


  SH assembler code partly based on x86 assembler code
  (c) Copyright 2002 - 2004 Marcus Comstedt (marcus@mc.pp.se) 

 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

/*******************************************************************************
  SNES9X for Mac OS (c) Copyright John Stiles
  
  Snes9x for Mac OS X

  (c) Copyright 2001 - 2004 zones
  (c) Copyright 2002 - 2004 107
  (c) Copyright 2002        PB1400c
  (c) Copyright 2004        Alexander, Sander and Steven Seeger
*******************************************************************************/

#include "port.h"

#include "mac-prefix.h"
#include "mac-os.h"
#include "mac-gworld.h"

static GDHandle		oldGD;
static GWorldPtr	oldGW;

void InitGWorld(GWorldPtr *world, const Rect *worldSize, short depth)
{
	QDErr	err;
	
	err = NewGWorld(world, depth, worldSize, nil, nil, 0L);
	if (err)
		QuitWithFatalError(err, "gworld 01");
	
	LockPixels(GetGWorldPixMap(*world));
	
	PrepareForGDrawing(*world);
	EraseRect(worldSize);
	FinishGDrawing(*world);
}

void PrepareForGDrawing(GWorldPtr world)
{
	GetGWorld(&oldGW, &oldGD);
	SetGWorld(world, nil);
}

void FinishGDrawing(GWorldPtr)
{
	SetGWorld(oldGW, oldGD);
}

Ptr SaveGWorldPixMap(GWorldPtr myWorld)
{
	PixMapHandle	GWorldPixMap;
	Ptr 			originalVideo;
	unsigned long	numRows, pixMapSize;
	bool			pixLocked;
	
	pixLocked = (GetPixelsState(GetGWorldPixMap(myWorld)) == pixelsLocked ? 1 : 0);
	
	LockPixels(GetGWorldPixMap(myWorld));
	
	GWorldPixMap = GetGWorldPixMap(myWorld);
	numRows = (*GWorldPixMap)->bounds.bottom - (*GWorldPixMap)->bounds.top;
	pixMapSize = GetPixRowBytes(GWorldPixMap) * numRows;
	originalVideo = NewPtr(pixMapSize);
	
	BlockMove(GetPixBaseAddr(GWorldPixMap), originalVideo, pixMapSize);
	
	if (!pixLocked)
		UnlockPixels(GetGWorldPixMap(myWorld));
	
	return originalVideo;
}

void RestoreGWorldPixMap(GWorldPtr myWorld, Ptr originalVideo)
{
	PixMapHandle	GWorldPixMap;
	unsigned long	numRows, pixMapSize;
	bool			pixLocked;
	
	pixLocked = (GetPixelsState(GetGWorldPixMap(myWorld)) == pixelsLocked ? 1 : 0);
	
	LockPixels(GetGWorldPixMap(myWorld));
	
	GWorldPixMap = GetGWorldPixMap(myWorld);
	numRows = (*GWorldPixMap)->bounds.bottom - (*GWorldPixMap)->bounds.top;
	pixMapSize = GetPixRowBytes(GWorldPixMap) * numRows;
	
	BlockMove(originalVideo, GetPixBaseAddr(GWorldPixMap), pixMapSize);
		
	if (!pixLocked)
		UnlockPixels(GetGWorldPixMap(myWorld));
}

void DrawSubCGImage(CGContextRef ctx, CGImageRef image, CGRect src, CGRect dst)
{
    float	w = (float) CGImageGetWidth(image);
    float	h = (float) CGImageGetHeight(image);

	CGRect	drawRect = CGRectMake(0.0, 0.0, w, h);

	if (!CGRectEqualToRect(src, dst))
	{ 
		float	sx = CGRectGetWidth(dst)  / CGRectGetWidth(src);
		float	sy = CGRectGetHeight(dst) / CGRectGetHeight(src);
		float	dx = CGRectGetMinX(dst) - (CGRectGetMinX(src) * sx);
		float	dy = CGRectGetMinY(dst) - (CGRectGetMinY(src) * sy);

		drawRect = CGRectMake(dx, dy, w * sx, h * sy);
	}

	CGContextSaveGState(ctx);
	CGContextClipToRect(ctx, dst);
	CGContextDrawImage(ctx, drawRect, image);
	CGContextRestoreGState(ctx);
}

CGImageRef CreateCGImageFromGWorld(GWorldPtr world)
{
	CGDataProviderRef	prov;
	CGColorSpaceRef		color;
	CGImageRef			image = nil;
	PixMapHandle		pixmap;
	Rect				bounds;
	long				rowbytes, w, h;
	char				*addr, *buf;
	
	if (!world)
		return (nil);
	
	pixmap = GetGWorldPixMap(world);
	if (GetPixDepth(pixmap) != 32)
		return (nil);

	rowbytes = GetPixRowBytes(pixmap);
	addr     = GetPixBaseAddr(pixmap);
	GetPixBounds(pixmap, &bounds);
	
	w = bounds.right  - bounds.left;
	h = bounds.bottom - bounds.top;
	
	buf = (char *) malloc(rowbytes);
	if (buf)
	{
		for (int i = 0; i < (h >> 1); i++)
		{
			memcpy(buf, addr + i * rowbytes, rowbytes);
			memcpy(addr + i * rowbytes, addr + (h - i - 1) * rowbytes, rowbytes);
			memcpy(addr + (h - i - 1) * rowbytes, buf, rowbytes);
		}

		free(buf);
	}

	prov = CGDataProviderCreateWithData(nil, addr, rowbytes * h, nil);
	if (prov)
	{	
		color = CGColorSpaceCreateDeviceRGB();
		if (color)
		{
			image = CGImageCreate(w, h, 8, 32, rowbytes, color, kCGImageAlphaNoneSkipFirst, prov, nil, 0, kCGRenderingIntentDefault);
			CGColorSpaceRelease(color);
		}
		
		CGDataProviderRelease(prov);
	}
	
	return (image);
}
